#include "HashMap.h"
#include <cassert>
#include <cmath>


template <class K, class V>
HashMap<K, V>::HashMap(int size, int (*hashFn)(const K &), Rehash probe = LINEAR, int (*doubleHashFn)(const K &) = DefaultDoubleHashFn ) {

	numItems = 0;
	hashFunc = hashFn;
	doubleHashFunc = doubleHashFn;
	rehashProbe = probe;
	if (probe == LINEAR) {
		tableSize = size;
	} else {
		// Rehashing strategies work with power of 2 table size, so make it the next larger 
		tableSize = static_cast<int>(pow(2.0,ceil(log(static_cast<double>(size))/log(static_cast<double>(2)))));
	} // end if
	table = new HashEntry<K, V> [tableSize];

	for (int i = 0; i < tableSize; i++) {
		table[i].slotState = EMPTY;
	} // end for
} // end HashMap constructor


template <class K, class V>
V HashMap<K, V>::get(K key) {

	int homeAddress = (abs(hashFunc(key)) % tableSize);
	int hashSlot = homeAddress;
	int numberOfRehashes = 0;
	int doubleHashingOffset;

	if (rehashProbe == DOUBLE) {
		doubleHashingOffset = abs(doubleHashFunc(key)) % tableSize;
		if (doubleHashingOffset % 2 == 0) {  // if it is even, make it odd
			doubleHashingOffset += 1;		
		} // end if
	} // end if


	while (table[hashSlot].slotState != EMPTY) {
		if (table[hashSlot].slotState == USED && table[hashSlot].key == key) {
			return table[hashSlot].value;
		} else { // either DELETED or wrong key
			numberOfRehashes++;
			assert(numberOfRehashes != tableSize);
			if (rehashProbe == LINEAR) {
				hashSlot = (hashSlot + 1) % tableSize;
			} else if (rehashProbe == QUADRATIC) {
				hashSlot = (homeAddress + (numberOfRehashes * numberOfRehashes + numberOfRehashes)/2) % tableSize;

			} else { // rehashProbe == DOUBLE
				hashSlot = (hashSlot + doubleHashingOffset) % tableSize;
			} // end if
		} // end if
	} // end while
	assert(table[hashSlot].slotState != EMPTY);

	return NULL; // should never get here
}


template <class K, class V>
bool HashMap<K, V>::contains(K key) {

	int homeAddress = (abs(hashFunc(key)) % tableSize);
	int hashSlot = homeAddress;
	int numberOfRehashes = 0;
	int doubleHashingOffset;

	if (rehashProbe == DOUBLE) {
		doubleHashingOffset = abs(doubleHashFunc(key)) % tableSize;
		if (doubleHashingOffset % 2 == 0) {  // if it is even, make it odd
			doubleHashingOffset += 1;		
		} // end if
	} // end if


	while (table[hashSlot].slotState != EMPTY) {
		if (table[hashSlot].slotState == USED && table[hashSlot].key == key) {
			return true;
		} else { // either DELETED or wrong key
			numberOfRehashes++;
			assert(numberOfRehashes != tableSize);
			if (rehashProbe == LINEAR) {
				hashSlot = (hashSlot + 1) % tableSize;
			} else if (rehashProbe == QUADRATIC) {
				hashSlot = (homeAddress + (numberOfRehashes * numberOfRehashes + numberOfRehashes)/2) % tableSize;
			} else { // rehashProbe == DOUBLE
				hashSlot = (hashSlot + doubleHashingOffset) % tableSize;
			} // end if
		} // end if
	} // end while
	return false;
}



template <class K, class V>
void HashMap<K, V>::put(K key, V value) {

	int homeAddress = (abs(hashFunc(key)) % tableSize);
	int hashSlot = homeAddress;
	int numberOfRehashes;
	int doubleHashingOffset;

	if (rehashProbe == DOUBLE) {
		doubleHashingOffset = abs(doubleHashFunc(key)) % tableSize;
		if (doubleHashingOffset % 2 == 0) {  // if it is even, make it odd
			doubleHashingOffset += 1;		
		} // end if
	} // end if

	numberOfRehashes = 0;
//	cout << "In put:  trying hashSlot " << hashSlot << endl;
	while (table[hashSlot].slotState != EMPTY && table[hashSlot].slotState != DELETED) {
		if (table[hashSlot].slotState == USED && table[hashSlot].key == key) {
			table[hashSlot].value = value;
		} else { // USED but wrong key
			numberOfRehashes++;
			assert(numberOfRehashes != tableSize);
			if (rehashProbe == LINEAR) {
				hashSlot = (hashSlot + 1) % tableSize;
			} else if (rehashProbe == QUADRATIC) {
				hashSlot = (homeAddress + (numberOfRehashes * numberOfRehashes + numberOfRehashes)/2) % tableSize;

			} else { // rehashProbe == DOUBLE
				hashSlot = (hashSlot + doubleHashingOffset) % tableSize;
			} // end if
		} // end if
//		cout << "In put:  trying hashSlot " << hashSlot << endl;

	} // end while
	table[hashSlot].key = key;
	table[hashSlot].value = value;
	table[hashSlot].slotState = USED;
	numItems++;

} // end put



template <class K, class V>
void HashMap<K, V>::remove(K key) {

	int homeAddress = (abs(hashFunc(key)) % tableSize);
	int hashSlot = homeAddress;
	int numberOfRehashes = 0;
	int doubleHashingOffset;

	if (rehashProbe == DOUBLE) {
		doubleHashingOffset = abs(doubleHashFunc(key)) % tableSize;
		if (doubleHashingOffset % 2 == 0) {  // if it is even, make it odd
			doubleHashingOffset += 1;		
		} // end if
	} // end if


	while (table[hashSlot].slotState != EMPTY) {
		if (table[hashSlot].slotState == USED && table[hashSlot].key == key) {
			table[hashSlot].slotState = DELETED;
			numItems--;
			return;
		} else { // either DELETED or wrong key
			numberOfRehashes++;
			assert(numberOfRehashes != tableSize);
			if (rehashProbe == LINEAR) {
				hashSlot = (hashSlot + 1) % tableSize;
			} else if (rehashProbe == QUADRATIC) {
				hashSlot = (homeAddress + (numberOfRehashes * numberOfRehashes + numberOfRehashes)/2) % tableSize;
			} else { // rehashProbe == DOUBLE
				hashSlot = (hashSlot + doubleHashingOffset) % tableSize;
			} // end if
		} // end if
	} // end while
	assert(table[hashSlot].slotState != EMPTY);

}  // end remove

template <class K, class V>
int HashMap<K, V>::size() const {
	return numItems;
} // end size

template <class K, class V>
int HashMap<K, V>::getHashTableSize() const {
	return tableSize;
} // end getHashTableSize


template <class K, class V>
double HashMap<K, V>::loadFactor() const {
	return double(numItems)/tableSize;
} // end size


template <class K, class V>
void HashMap<K, V>::clear() {

	for (int i = 0; i < tableSize; i++) {
		table[i].slotState = EMPTY;
	} // end for

	numItems = 0;
} // end clear



template <class K, class V>
HashMap<K, V>::~HashMap() {
	delete[] table;
}  // end ~HashMap


template <class K, class V>
void HashMap<K, V>::display() {

	for (int i = 0; i < tableSize; i++) {
		cout << i <<": ";
		if (table[i].slotState == EMPTY) {
			cout << "(EMPTY Slot)" << endl;
		} else if (table[i].slotState == DELETED) {
			cout << "(DELETED Slot)" << endl;
		} else {
			cout << "Key = " << table[i].key << "      Value = " << table[i].value << endl;
		} // end if

	} // end for

} // end display

